package chapman;

import game.Game;
//import game.Strategy;

import java.util.ArrayDeque;
import java.util.Deque;
//import java.util.HashMap;
//import java.util.Iterator;
//import java.util.Map;
import java.util.Random;
import java.util.Vector;

import strategy.UtilityBasedStrategy;

import lemonade.GeneralizedLemonadeUtility;

/**
 * @author Archie C, Enrique MdC, Adam S.
 * 
 * GTTA = Game Theoretic TActics
 * 
 * This algorithm uses a refinement to the set of NE to construct a set of "tactics", 
 * which correspond to a set of "ideal types" that opponents can take and, correspondingly, 
 * are used to inform action selection.  Specifically, the leader-follower framework, 
 * successfully employed in the EA^2 and TeamUP algorithms for previous LSG tournaments, 
 * has been extended to include a new axis of types, namely, greedy, fair and generous 
 * opponents.  These new classifications are applied to leaders only (followers continue to 
 * be classified according to who they follow) giving a set of 6 types an agent can be 
 * labelled -- generousLeader, fairLeader, greedLeader, meFollower, otherFollower and Unknown 
 * -- so the GTTA algorithm has to reason over 36 joint types.  Much like our previous LSG
 * entries, ideal behaviour is posited as following the types listed above, and various 
 * measures of the deviations from these types are used to classify an opponent as one type 
 * or the other.
 * 
 * Since the nature of the LSG utility function has changed, GTTA employs new methods for 
 * measuring the stickIndex and followIndex of an opponent and for evaluating the usefulness 
 * of different actions. Most importantly, as the action configurations that would lead to 
 * beneficial cooperation between two players in the new LSG are not anywhere near as obvious 
 * as in previous iterations of the tournament, GTTA begins by fearing the worst -- that its 
 * opponents are, together, the perfect adversary.  As such, in the first rounds of the game 
 * GTTA employs a no regret algorithm for multi-armed bandits with complete monitoring, based 
 * on Exp3.  These rounds are used to learn the opponents' nature and test our domain-
 * specific methods for initiating collusion will be useful in future rounds.
  * 
 * In more detail, when assessing the collusive potential of an opponent, actions are only 
 * considered useful if they fall within a subset of the NE that we term the "winning 
 * coalition member (WCM) set" of the NE.  These are the NE in which GTTA earns greater than 
 * at least one of the other players (the is the NE refinement mentioned above).  The value 
 * of a joint action to GTTA is only incorporated into its reasoning if the joint action is 
 * in the set of WCM NE for GTTA.  Thus, implicit in the algorithm's motivation is the 
 * assumption that other algorithms trying to cooperate with GTTA will do so in a manner that 
 * is sufficiently close to one of the 6 ideal types above such that its actions are 
 * predictable to GTTA; that is we assume that other entrants will attempt to coordinate 
 * around the set of one--shot NE (as in previous LSG tournaments).  However, if these 
 * opponent models and tactics do not prove to be beneficial, the algorithm reverts to its 
 * worst-case assumptions.  
 * 
 * Furthermore, note that GTTA is a work in progress.  For example, it does not include the 
 * MDP-based reasoning of TeamUP, and instead relies on a simpler myopic action values (with 
 * the obvious flaws).  Similarly, a more principled measure of the distance of an opponent 
 * to an ideal type would likely improve GTTA, and also EA^2 and TeamUP.  Finally, there are 
 * many unused methods and variable type-outs, which are intended to be built upon in the 
 * future  
 */
public class GTTA extends UtilityBasedStrategy {

	GeneralizedLemonadeUtility utility;
	
	private Vector<Integer> mySequence = new  Vector<Integer>() ;
	private Agent opponents[] = new Agent[2] ;
	public Tactics gttaTactics ; // = new Tactics((GeneralizedLemonadeUtility)game.getUtility()) ;
	private int[][][] psNE ; // = gttaTactics.getPSNE() ;
	private ExpWeighting safetyStrategy ; // = new Exp3(numActions) ;
//	private double[][][] expReward ; // = gttaTactics.getInitialRewards() ;
	private int currentAction ;
	private int exploitAction ;
	private int safeAction ;
	private double[] avgRewards = new double[2];
	public int timeStep ;
	private static final int INIT_SEARCH = 10 ; 

	/**
	 * @param g
	 * @param seed
	 */
	public GTTA(Game game, long seed, String options) {
		super(game, seed) ;
		timeStep = 0 ;
		this.utility = (GeneralizedLemonadeUtility)game.getUtility();
		
		// Generate opponent models
		for (int i = 0; i < game.getNumPlayers()-1; i++) {
			opponents[i] = new Agent(numActions) ;
		}
				
		// Generate tactics
		gttaTactics = new Tactics(utility) ;
		psNE = gttaTactics.getPSNE() ;
		safetyStrategy = new ExpWeighting(utility) ;
		
		// Initialise expected reward for state-tactics pairs (not done), and average rewards for strategy components
//		expReward = gttaTactics.getInitialExpRewards() ;
		avgRewards[0] = 0.0 ;
		avgRewards[1] = 0.0 ;
		
		// First move = take greedyLeader position
		int[] initialActions = gttaTactics.getGreedyActions() ;
		currentAction = initialActions[r.nextInt(initialActions.length)];		
	}

	
	public void observeOutcome(int[] actions) {
		
		// Store actions
		currentAction = actions[0];
		mySequence.add(currentAction);
		for (int i = 0; i < opponents.length; i++) 
			opponents[i].lastAction(actions[i+1], actions[0]) ;
		
		// Compute strategy components
		safetyStrategy.updateExpWeighting(actions, utility) ;
		exploitAction = computeStrat(actions) ;
		safeAction = safetyStrategy.getExpWeightingAction() ;
		
		//Choose a strategy: exploit or safety
		updateAvgRewards(actions) ;
		if (timeStep > INIT_SEARCH){
			if (avgRewards[0] > avgRewards[1]){
				currentAction = exploitAction ;
			} 
			else {  
				currentAction = safeAction ;
			}
		}
				
		timeStep++;
 
	}

	@Override
	public void observeUtilities(int[] possibleUtilities, int currentAction) {
	}
	
	@Override
	public int getAction() { 
		return currentAction;	
	}
	
	/**
	 * Computes current state, i.e. Maps from measures to opponents' current joint strategy in terms of ideal types.
	 * Use to get six ideal types: genL, fairL and greedyL, meF, otherF and Unknown/Random
	 */
	private int[][] computeTypes(){
		// decision flow for a type feature
		int[][] type = new int[opponents.length][2] ; 
		for (int i = 0; i < opponents.length; i++) {
			int lfType = 0;
			int gfgType = 1;
			int j;
			if (i == 0) j = 1;
			else j = 0;
			
			// leaderFollower axis
			double fOp = opponents[i].followIndex(opponents[j].getSequence(), psNE);
			double fMe = opponents[i].followIndex(mySequence, psNE);
			//System.out.println("followIndex:"+ fOp+" threshold:"+opponents[i].threashold() );
			//System.out.println("followIndex:"+ fOp+" threshold:"+opponents[i].threashold() );
			if(fMe < opponents[i].lfThreashold()){
				lfType = 1 ; // is following me
			}
			else if(fOp < opponents[i].lfThreashold()){
				lfType = 2 ; // is following other opponent
			}
			else if(opponents[i].isSticking()){
				lfType = 3 ; // is a leader
			} // else{ lfType = 0; // is not predictable == unknown, random }
			
			// greedyfairGenerous axis 
			double gfg = opponents[i].gfgIndex(gttaTactics.getGfgMeasure()) ; 
			//System.out.println("gfgIndex: "+ gfg +"; threshold: "+ opponents[i].gfgThreshold() );
			if (gfg < - opponents[i].gfgThreashold() ) {
				gfgType = 0 ; // if gfg < -B, i is labelled "generous" 
			} else if ( gfg > opponents[i].gfgThreashold() ){
				gfgType = 2 ; // if -B <= gfg <= B, i is labelled "greedy" 
			} // else if -B <= gfg <= B, i is labelled "fair"
			type[i][0] = lfType;  // 0 random; 1 follows me; 2 follows other; 3 leader
			type[i][1] = gfgType; // 0 generous; 1 fair; 2 greedy
		}
		return type ;
	}
		
	/**
	 * Computes an action to take in the current state, given a context (current joint action profile).
	 * @return currentAction 
	 * TODO: VI. Currently only myopic action selection -- no planning, no back-propagation.
	 */
	private int computeStrat(int[] actions) {
		int gttaAction = currentAction;
		int[][] currentTypes = computeTypes() ;//	type[i][0] = lfType, type[i][1] = gfgType
		int[] context = {actions[1], actions[2]} ; // others' actions is the context

		// type[i][0] = lfType;  // 0 random; 1 follows me; 2 follows other; 3 leader
		// type[i][1] = gfgType; // 0 generous; 1 fair; 2 greedy

		int[] typeActs = new int[7] ;  
		// 0. greedy lead
		int[] greedyActs = gttaTactics.getGreedyActions() ;
		typeActs[0] = greedyActs[r.nextInt(greedyActs.length)];
		// 1. fair lead
		int[] fairActs = gttaTactics.getFairActions() ;
		typeActs[1] = fairActs[r.nextInt(fairActs.length)];
		// 2. generous lead			
		int[] generousActs = gttaTactics.getGenerousActions() ;
		typeActs[2] = greedyActs[r.nextInt(generousActs.length)];
		// 3. follow_0			
		int[] follow_0Acts = gttaTactics.optResponseToJ(context[0], utility) ;
		if (follow_0Acts != null){
			typeActs[3] = follow_0Acts[r.nextInt(follow_0Acts.length)];
		} else {
			int[] newActs = gttaTactics.getFairActions() ;
			typeActs[3] = newActs[r.nextInt(newActs.length)];
		} 
		// 4. follow_1
		int[] follow_1Acts = gttaTactics.optResponseToJ(context[1], utility) ;
		if (follow_1Acts != null){
			typeActs[4] = follow_1Acts[r.nextInt(follow_1Acts.length)];
		} else {
			int[] newActs = gttaTactics.getFairActions() ;
			typeActs[4] = newActs[r.nextInt(newActs.length)];
		} 
		// 5. default to safety			
		typeActs[5] = safetyStrategy.getExpWeightingAction();
		
		// 6. play best response
		int[] reward = new int[numActions]; 
		int maxReward = 0;
		for(int ai =0; ai< numActions ; ai++ ){
			int util[] = utility.getUtility(new int[]{ai,actions[1], actions[2]}) ;
			reward[ai] = util[0];
			if (reward[ai] > maxReward)
				maxReward = reward[ai] ;
		}
		int bestNum = -1 ;
		int[] best = new int[numActions] ;
		for(int ai =0; ai< numActions ; ai++ ){
			if(maxReward == reward[ai]){
				bestNum++ ;
				best[bestNum] = ai ;
			}
		}
		int[] newActs = new int[bestNum + 1] ; 
		for(int j =0; j< bestNum ; j++ ){
			newActs[j] = best[j] ;
		}
		typeActs[6] = newActs[r.nextInt(newActs.length)];
		
		/** 
		 * Begin the useful cases
		 */
		// type[i][0] = lfType;  // 0 random; 1 follows me; 2 follows other; 3 leader
		// type[i][1] = gfgType; // 0 generous; 1 fair; 2 greedy
		
		// if one or the other is following me
		if (currentTypes[0][0] == 1 || currentTypes[1][0] == 1){ 		
			gttaAction = typeActs[0] ;
		}
		
		// If one or the other is sticking, but not both, 
		else if (currentTypes[0][0] == 3 && !(currentTypes[1][0] == 3)) {
			if ( !(currentTypes[0][1] == 2) ) // ...and it is not too greedy
				gttaAction = typeActs[3] ;
			else // ...but the sticker is greedy
				gttaAction = typeActs[1] ;			
		}
		else if (!(currentTypes[0][0] == 3) && currentTypes[1][0] == 3) {
			if ( !(currentTypes[1][1] == 2) )
				gttaAction = typeActs[4] ;
			else
				gttaAction = typeActs[1] ;
		}
				
		// If both are sticking,  // type[i][1] = gfgType;
		else if (currentTypes[0][0] == 3 && currentTypes[1][0] == 3){ 
			if (currentTypes[0][1] == 2 && !(currentTypes[1][1] == 2)){
				// ...follow more generous one,
				gttaAction = typeActs[4] ;	
			}
			else if (!(currentTypes[0][1] == 2) && currentTypes[1][1] == 2){
				gttaAction = typeActs[3] ;
			} 
			else { // ...or if both are generous
				gttaAction = typeActs[6] ; // best response
			}
			
		}	
		else {
			gttaAction = typeActs[5] ;
		}
		return gttaAction ;
	}
	
	private void updateAvgRewards(int[] actions) {
		final double ALPHACONST = 1 ;
		final double ALPHARHO = 1/2 ; // ALPHARHO = 1 gives mean, ALPHARHO -> 0 gives more weight to recent observations  
		int[] util ;
		int[] stratReward = new int[2] ;
		util = utility.getUtility(new int[]{exploitAction , actions[1], actions[2] });
		stratReward[0] = util[0] ;
		util = utility.getUtility(new int[]{safeAction , actions[1], actions[2] });		
		stratReward[1] = util[0] ;
		double alpha = Math.pow( ALPHACONST/(timeStep + 1), ALPHARHO );
		for (int k = 0 ; k < 2 ; k++ ){
			avgRewards[k] = (1-alpha) * avgRewards[0]/ + alpha * stratReward[k] ;
		}
	}
}

/**
 * @author Enrique, Archie
 */
class Agent{
	int numActions;
	// private double stick = 6; //values range from 0 to 6
	//private Map<Boolean, Double> followIndex = new HashMap<Boolean, Double>();
	private int strat[];
	private Vector<Integer> sequence = new  Vector<Integer>();
	private Deque<Integer> pastActions;
	private static final int CAPACITY = 20; 
	private static final double GAMMA = 0.8; // discounts past actions
	private static final double RHO = 0.6 ; // scales distances from ideal play
	private static final double TRESHOLD_LF = (Math.pow(6, RHO)/(1-GAMMA))*0.3; //this is 0.3 from max followIndex, anything bellow this is considered follower
	private static final double TRESHOLD_GFG = 0.6 ; 
	// dividing [-1,1] into three regions [-1, -thres], [-thre,thres], [thres, 1]
	// the smaller the value, the more likely to label as greedy or generous 
	private int myLastActions[];
	

	protected Agent(int actions){
		this.numActions = actions;
		strat = new int[numActions];
		myLastActions = new int[2];
		java.util.Arrays.fill(strat,0);
		pastActions = new ArrayDeque<Integer>(CAPACITY);
		java.util.Arrays.fill(myLastActions,-1);
	}

	protected void lastAction(int hisAction, int myAction){
		strat[hisAction]++;
		sequence.add(hisAction);
		if(pastActions.size() >= CAPACITY)
			pastActions.removeLast();
		pastActions.addFirst(hisAction);
		myLastActions[1] = myLastActions[0];
		myLastActions[0] = myAction;
	}

	/**
	 * computes the followIndex of a given sequence
	 * @param opponentSeq the sequence to compare to
	 * @return the normalised followIndex of the its sequence and the argument's sequence
	 * TODO: replace these indices with better model selection framework 
	 */
	protected double followIndex(Vector<Integer> mySeq, Vector<Integer> otherSeq, int[][][] psNE){
		double fIndex = 0.0;
		double sumGAMMA = 0.0;
		double dif;
		double difMe;
		double difOther;
		double norm;
		int t = 0;
		assert(this.sequence.size() == mySeq.size() && this.sequence.size() == otherSeq.size());
		for (int i = 0; i < sequence.size(); i++) 
			sumGAMMA += Math.pow(GAMMA, i);

		for(int i = sequence.size()-1 ; i >0 ; i--) {
			int[] idealAiMe = idealFollowerActions(mySeq.get(i-1), psNE) ;
			int[] idealAiOther = idealFollowerActions(otherSeq.get(i-1), psNE) ;
			//
			if (idealAiMe != null ) {
				int[] dist = new int[idealAiMe.length] ;
				int minDist = numActions/2 ;
				for (int l = 0 ; l < idealAiMe.length ; l++ ){
					dist[l] = Math.abs((sequence.get(i) - idealAiMe[l]) % numActions) ;
					if(dist[l] > numActions/2) dist[l] = numActions - dist[l] ;
					if (dist[l] < minDist) minDist = dist[l] ;
				}
				difMe = minDist ;
			} else difMe = numActions ;
			//
			if (idealAiOther != null ) {
				int[] dist = new int[idealAiOther.length] ;
				int minDist = 6 ;
				for (int l = 0 ; l < idealAiOther.length ; l++ ){
					dist[l] = Math.abs((sequence.get(i) - idealAiOther[l]) % numActions) ;
					if(dist[l] > 6) dist[l] = numActions - dist[l] ;
					if (dist[l] < minDist) minDist = dist[l] ;
				}
				difOther = minDist ;
			} else difOther = numActions ;
			//
			dif = Math.min(difMe, difOther);
			if (dif == numActions) dif = 0;
			if(t==0) dif = Math.pow(dif, RHO);
			norm = Math.pow(GAMMA, t) / sumGAMMA;
			fIndex += norm * dif ;
			t++;
		}
		fIndex = - fIndex;
		return fIndex ;
	}

	/**
	 * computes the followIndex of a given sequence for a given opponent
	 * @param opponentSeq the sequence to compare to
	 * @return the normalised followIndex of the its sequence and the argument's sequence
	 */
	protected double followIndex(Vector<Integer> otherSeq, int[][][] psNE){
		double fIndex = 0;
		double sumGAMMA = 0;
		double dif;
		double norm;
		int t = 0;
		assert(this.sequence.size() ==  otherSeq.size());
		for (int i = 0 ; i < sequence.size() ; i++) 
			sumGAMMA += Math.pow(GAMMA, i) ;

		for(int i = sequence.size()-1 ; i >0 ; i--) {
			int[] idealAi = idealFollowerActions(otherSeq.get(i-1), psNE) ;
			if (idealAi != null ) {
				int[] dist = new int[idealAi.length] ;
				int minDist = 6 ;
				for (int l = 0 ; l < idealAi.length ; l++ ){
					dist[l] = Math.abs((sequence.get(i) - idealAi[l]) % 12) ;
					if(dist[l] > 6)
						dist[l] = 12 - dist[l] ;
					if (dist[l] < minDist)
						minDist = dist[l] ;
				}
				dif = minDist ;
				if(t==0) dif = Math.pow(dif, RHO);
				norm = Math.pow(GAMMA, t) / sumGAMMA;
				fIndex += norm * dif ;
				t++;
			}
		}
		return fIndex ;
	}

	/**
	 * computes an opponent's stickIndex
	 * @return the normalised stickiness value, 0 is an ideal leader 
	 */
	protected double stickIndex(){
		double sIndex = 0;
		double sumG = 0;
		double dif;
		double norm;
		int t = 0;
		for (int i = 0; i < sequence.size(); i++) 
			sumG += Math.pow(GAMMA, i);

		for(int i = sequence.size()-1; i >0 ; i--) {
			dif = Math.min(Math.abs(sequence.get(i) - sequence.get(i-1)), 
					Math.abs(sequence.get(i-1) - sequence.get(i)));
			if(dif > 6)
				dif = 12 - dif;
			if(t==0) 
				dif = Math.pow(dif,RHO);
			norm = Math.pow(GAMMA, t) / sumG;
			sIndex += norm * dif;
			t++;
		}
		sIndex = - sIndex;
		return sIndex;
	}
	
	public boolean isSticking(){
		return(stickIndex() < TRESHOLD_LF);
	}
	

	/**
	 * Computes its gfgIndex, which measures the relative difference in surplus earned by colluders
	 * @return the gfg value, -1 is an perfectly generous, 1 is perfectly greedy
	 */
	protected double gfgIndex(double[] gfgMeasure) {
		double gIndex = 0;
		double sumGAMMA = 0;
		double norm;
		int t = 0;
		for (int i = 0; i < sequence.size(); i++) 
			sumGAMMA += Math.pow(GAMMA, i);
		for(int i = sequence.size()-1; i >0 ; i--) {
			double gfgValue = gfgMeasure[sequence.get(i)] ; 
			norm = Math.pow(GAMMA, t) / sumGAMMA;
			gIndex += norm * gfgValue;
			t++;
		}
		return gIndex;
	}
	
	
	protected Vector<Integer> getSequence(){
		return sequence;
	}
	

	protected int currentAction(){
		return pastActions.getFirst();
	}
	
	/**
	 * Generates the actions that an ideal follower may take. 
	 * No reasoning about selection of equilibrium here, only existence of NE.
	 * @return idealActions a set of actions consistent with ideal NE follower model 
	 * @param ai my action
	 * @param psNE the joint action space with psNE = 1, non-psNE =0
	 */
	protected int[] idealFollowerActions(int ai, int[][][] psNE ) {
		int actNum = -1;
		int[] idealActs = new int[numActions] ;
		for (int aj = 0 ; aj < numActions ; aj++ ){
			boolean isPSNE = false ;
			int ak = 0 ;
			while (isPSNE == false && ak < numActions) {
				if (psNE[ai][aj][ak] == 1) isPSNE = true ;
				ak++ ;
			}
			if (isPSNE == true ) {
				actNum++ ;
				idealActs[actNum] = aj ;
			} 
		}
		if (actNum > -1) {
			int[] idealActions = new int[actNum + 1];
			for (int i = 0 ; i <= actNum ; i++ ){
				idealActions[i] = idealActs[i] ;
			}
			assert (idealActs.length <= numActions ) ;
			assert (idealActs.length == actNum -1 ) ;
			return idealActions ;
		}
		else { 
			assert(idealActs[0] == -1) ;
			return null ;
		}
	}
	public double lfThreashold(){
		return TRESHOLD_LF;
	}
	public double gfgThreashold(){
		return TRESHOLD_GFG;
	}
}



/**
 * Generate tactics according to simple stability concepts: Pure-strategy Nash equilibria with refinements.  
 * Ideal types follow certain tactics. 
 * Deviations of actual play from ideal types then informs the action for GTTA to take.
 * @author Archie
 */
class Tactics{

	GeneralizedLemonadeUtility utility ;
	Random r = new Random() ;
	
	private int numActions;
	private static int[][][] psNE ;
	private static int[][][] difNE ;
//	private static int[][] winningCMSet;
	private static double[] gfgMeasure ;
	private static double[] leadQuality ;
	private static int[] greedyLeads ;
	private static int[] fairLeads ;
	private static int[] generousLeads ;


	public Tactics(GeneralizedLemonadeUtility utility) {
		numActions = utility.getIslandSize() ;
		psNE = computePsne(utility) ;
		difNE = computeDiff(utility) ;
//		winningCMSet = computeWcmSet(utility) ;
		gfgMeasure = computeGfgMeasure(utility) ;
		leadQuality = computeLeadQuality(utility);
		greedyLeads = setGreedyLeads() ;
		fairLeads = setFairLeads() ;
		generousLeads = setGenerousLeads() ;

	} 

	/**
	 * Compute all pure strategy NE; these are the backbone of the strategy, 
	 * and are used because they provide stability and predictability over the short run.
	 * @return psne A multi-dimensional array indicating if {ai,aj,ak} is a NE.
	 */
	protected int[][][] computePsne(GeneralizedLemonadeUtility utility){
		int[][][] util1 = new int[numActions][numActions][numActions] ; 
		int[][][] bestR1 = new int[numActions][numActions][numActions] ; 
		int[][][] psne = new int[numActions][numActions][numActions] ;

		for (int a3 = 0; a3 < numActions ; a3++){
			for (int a2 = 0 ; a2 < numActions ; a2++){
				int[] contextUtil1 = new int[numActions] ;	// use this later to compute bestResponse
				for (int a1 = 0 ; a1 < numActions ; a1++){
					int[] util = utility.getUtility(new int[]{a1,a2,a3}) ;
					util1[a1][a2][a3] = util[0] ;
					contextUtil1[a1] = util[0] ;
				}
				int maxValue = contextUtil1[0] ;
				for(int a1=1 ; a1<numActions ; a1++){  
					if(contextUtil1[a1] > maxValue){  
						maxValue = contextUtil1[a1];  
					}  
				}  
				for(int a1=0 ; a1<numActions ; a1++){  
					if(contextUtil1[a1] == maxValue){  
						bestR1[a1][a2][a3] = 1 ;  
					}  
				} 
			}
		}		

		for (int a3 = 0; a3 < numActions ; a3++){
			for (int a2 = 0 ; a2 < numActions ; a2++){
				for (int a1 = 0 ; a1 < numActions ; a1++){
					//					Can do this because utilities are symmetric
					psne[a1][a2][a3] = bestR1[a1][a2][a3] * bestR1[a2][a1][a3] * bestR1[a3][a2][a1];
				}
			}
		}
		return psne ;
	}
	
	public boolean isPSNE(int ai, int aj, int ak){
		if (psNE[ai][aj][ak] == 1) 
			return true ; 
		else 
			return false ;
	}
	
	public int[][][] getPSNE(){ 
		return psNE ; 
		}		
	
	/**
	 * Compute the payoff difference between two actions for each pure strategy NE; 
	 * These differences guide the choice of action when cooperating with a particular player
	 * (i.e. when and how to be greedy, fair or generous)
	 * @return diff A multi-dimensional array of the differences 
	 * between utility_j and utility_k given a fixed a_i
	 */
	protected int[][][] computeDiff(GeneralizedLemonadeUtility utility) {
		int[][][] diff = new int[numActions][numActions][numActions] ;

		for (int ak = 0; ak < numActions ; ak++){
			for (int aj = 0 ; aj < numActions ; aj++){
				for (int ai = 0 ; ai < numActions ; ai++){ 
					int util_j[] = utility.getUtility(new int[]{aj,ai,ak}) ; 
					int util_k[] = utility.getUtility(new int[]{ak,aj,ai}) ; 
					diff[ai][aj][ak] = psNE[ai][aj][ak] * ( util_j[0] - util_k[0] ) ;
				}
			}
		}
		return diff ;
	}
	
	public int[] getDifNE(int ai, int aj, int ak){
		int[] difference = { 0 };
		if (isPSNE(ai,aj,ak) == true ){
			difference[0] = difNE[ai][aj][ak] ;
			return difference ;
		}
		else return null;
	}	
	
	/**
	 * Computes the subset of NE in which i is in the winning coalition;
	 * Effectively a refinement that selects good NE.
	 * @return wcmSet A list of joint actions that make i a winning coalition member, 
	 * and the corresponding utilities 
	 */
	protected int[][] computeWcmSet(GeneralizedLemonadeUtility utility){
		int[][] wcmPre = new int[numActions*numActions*numActions][numActions] ;
		int numSet= -1;
		for (int ak = 0; ak < numActions ; ak++){
			for (int aj = 0 ; aj < numActions ; aj++){
				for (int ai = 0 ; ai < numActions ; ai++){ 
					if (isInWcmSet(ai,aj,ak)){
						numSet++ ;
						wcmPre[numSet][1] = ai ; 
						wcmPre[numSet][2] = aj ;
						wcmPre[numSet][3] = ak ;
					}
				}
			}
		}
		if (numSet > 0){
			int[][] wcmSet = new int[numSet+1][numActions] ;
			for (int l = 0; l < numSet ; l++){
				wcmSet[numSet][1] = wcmPre[numSet][1] ;
				wcmSet[numSet][2] = wcmPre[numSet][2] ;
				wcmSet[numSet][3] = wcmPre[numSet][3] ;
			}
			return wcmSet ;
		} 
		else return null ;
	}
	
	public boolean isInWcmSet(int ai, int aj, int ak){
		if (isPSNE(ai,aj,ak)){
			int[] dif1 = getDifNE(ak,ai,aj) ;
			int[] dif2 = getDifNE(aj,ai,ak) ;
			int utilDiff1 = dif1[0] ;
			int utilDiff2 = dif2[0] ;
			boolean isInWCM = false ;
			if (utilDiff1>0 || utilDiff2>0 )
				isInWCM = true ;
			return isInWCM ;
		} 
		else return false ;
	}
	
	/**
	 * This measures the greed-fair-generosity of an opponent's lead.
	 * It is based on the belief that GTTA's opponents expect equilibrium play from their opponents,
	 * and that they also expect an equilibrium to be selected at random.
	 * At some stage this c/should be superseded by some higher k-level reasoning.
	 * @return gfgMeas The gfmMeaure for each action, on a scale of -1 to 1: greedy = 1, fair = 0, generous = -1
	 */
	protected double[] computeGfgMeasure(GeneralizedLemonadeUtility utility){
		double[] gfgMeas = new double[numActions] ;
		for (int a1 = 0 ; a1 < numActions ; a1++ ) {
			double expUtil = 0 ;
			int numNE = 0 ; 
			for (int a2 = 0 ; a2 < numActions ; a2++ ) {
				for (int a3 = 0 ; a3 < numActions ; a3++ ){
					if (isPSNE(a1,a2,a3) == true){
						numNE++ ;
						int[] util = utility.getUtility(new int[]{a1,a2,a3}) ; 
						// Expected reward against two randomly selected psNe-playing opponents 
						expUtil = ( (numNE-1)* expUtil / numNE ) + util[0]/numNE ;
					}
				}
			}
			gfgMeas[a1] = expUtil ;
		}
		// Normalise
		double maxValue = 0 ;
		double minValue = 0 ;
		for(int a1=0 ; a1<numActions ; a1++){  
			if(gfgMeas[a1] > maxValue){  
				maxValue = gfgMeas[a1] ;  
			}  else if (gfgMeas[a1] < minValue){
				minValue = gfgMeas[a1] ;
			}
		}
		// On a scale of -1 to 1: greedy = -1, fair = 0, generous = 1
		for(int a1=0 ; a1<numActions ; a1++){  
			gfgMeas[a1] = (2*gfgMeas[a1] - (minValue + maxValue))/ (minValue - maxValue)  ;
		}
		return gfgMeas ; 
	}
	
	public double[] getGfgMeasure(){
		return gfgMeasure ;
	}	
	
	/** This section computes the best lead strategies for the different gfg types GTTA can take.
	 * First method computes the leadQuality, which is my expected surplus division within the set of WCM NE.
	 * Following three methods compute the greedy, fair and generous lead actions, which maximise, zero and 
	 * minimise my expSurplusDiv, respectively.  
	 */
	
	/**
	 * @return actionQuality The expSurplusDiv for each action, on a scale of -1 to 1: greedy = 1, fair = 0, generous = -1
	 */
		protected double[] computeLeadQuality(GeneralizedLemonadeUtility utility){
		double[] actionQuality = new double[numActions] ;
		for (int ai = 0 ; ai < numActions ; ai++ ) {
			double expSurplusDiv = 0 ;
			int numWcmNE = 0 ; 
			for (int aj = 0 ; aj < numActions ; aj++ ) {
				for (int ak = 0 ; ak < numActions ; ak++ ){
					double surplus = 0 ;
					double surplusDivision ;
					if (isInWcmSet(ai,aj,ak) == true){
						double difThem = difNE[ai][aj][ak];
						if (difThem > 0)
							surplus = difNE[aj][ai][ak] + difThem;
						else
							surplus = difNE[ak][ai][aj] - difThem;
						surplusDivision = difNE[ak][ai][aj]/surplus ;
						numWcmNE++ ;
						// Expected surplusDivision against at least one wcmNE-playing opponent 
						expSurplusDiv = ( (numWcmNE-1)* expSurplusDiv / numWcmNE ) + surplusDivision/numWcmNE ;
					}
				}
			}	
			actionQuality[ai] = expSurplusDiv ;
		} 
		return actionQuality ;
	}
	
	/**
	 * @return greedyActions The set of greediest actions
	 */
	protected int[] setGreedyLeads(){
		double[] actionQuality = leadQuality ;
		double maxQual = actionQuality[0];
		for (int aMe = 1; aMe < numActions ; aMe++ ){
			if (actionQuality[aMe] > maxQual)
				maxQual = actionQuality[aMe] ;
		}
		int[] greedyActs = new int[numActions] ;
		int numActs = 0 ;
		for (int aMe = 0 ; aMe < numActions ; aMe++){
			if (actionQuality[aMe] == maxQual)
				greedyActs[numActs] = aMe ;
				numActs++ ;
		}
		int[] greedyActions = new int[numActs] ;
		for (int j = 0 ; j < numActs ; j++){
			greedyActions[j] = greedyActs[j] ;
		}
		return greedyActions ;
	}
	
	/**
	 * @return generousActions The set of most generous actions
	 */	
	protected int[] setGenerousLeads(){
		double[] actionQuality = leadQuality  ;
		double minQual = actionQuality[0];
		for (int aMe = 1; aMe < numActions ; aMe++ ){
			if (actionQuality[aMe] < minQual)
				minQual = actionQuality[aMe] ;
		}
		int[] generousActs = new int[numActions] ;
		int numActs = 0 ;
		for (int aMe = 0 ; aMe < numActions ; aMe++){
			if (actionQuality[aMe] == minQual)
				generousActs[numActs] = aMe ;
				numActs++ ;
		}
		int[] generousActions = new int[numActs] ;
		for (int j = 0 ; j < numActs ; j++){
			generousActions[j] = generousActs[j] ;
		}
		return generousActions ;
	}
	
	/**
	 * @return fairActions The set of fairest actions
	 */	
	protected int[] setFairLeads(){
		double[] actionQuality =  leadQuality ;
		double nearestZeroQual = Math.abs(actionQuality[0]);
		for (int aMe = 1; aMe < numActions ; aMe++ ){
			if (Math.abs(actionQuality[aMe]) < nearestZeroQual)
				nearestZeroQual = actionQuality[aMe] ;
		}
		int[] fairActs = new int[numActions] ;
		int numActs = 0 ;
		for (int aMe = 0 ; aMe < numActions ; aMe++){
			if (actionQuality[aMe] == nearestZeroQual)
				fairActs[numActs] = aMe ;
				numActs++ ;
		}
		int[] fairActions = new int[numActs] ;
		for (int j = 0 ; j < numActs ; j++){
			fairActions[j] = fairActs[j] ;
		}
		return fairActions ;
	}
	
	public int[] getGreedyActions(){
		return greedyLeads ;
	}
	
	public int[] getFairActions(){
		return fairLeads ;
	}
	
	public int[] getGenerousActions(){
		return generousLeads ;
	}
	
	/**
	 * This method computes the follower tactics; that is, the best response to aj 
	 * assuming that the third player plays, or quickly converges, to a NE 
	 * @param aj
	 * @return optimalActions The optimal NE-consistent responses to aj, assuming aOther is randomly chosen
	 * @return null if no NE-consistent action exists.
	 * TODO: k-level reasoning for third player
	 */
	public int[] optResponseToJ(int aj, GeneralizedLemonadeUtility utility){
		int[] optimalActs = new int[numActions] ;
		double[] expUtil = new double[numActions] ;
		double maxExpUtil = 0 ;
		
		for (int aMe = 0 ; aMe < numActions ; aMe++ ) {
			double expValue = 0 ;
			int numNE = 0 ; 
			for (int aOther = 0 ; aOther < numActions ; aOther++ ) {
//	DOES BETTER NOT ASSUMING NE RESPONSES - Real potential for some higher k-level reasoning about the third player here.
//				if (isPSNE(aj,aMe,aOther) == true){ // if it is a NE
					numNE++ ;
					int[] util = utility.getUtility(new int[]{aMe,aj,aOther}) ;
					// Expected reward against a r psNe-playing opponent and a random
					expValue = ( (numNE-1)* expValue / numNE ) + util[0]/numNE ;
//				}
			}
			
			expUtil[aMe] =  expValue ;
			if(expUtil[aMe] > maxExpUtil) 
				maxExpUtil = expUtil[aMe] ;
		}
		int optActNum = -1 ;
		for (int aMe = 0 ; aMe < numActions ; aMe++ ) {
			if (expUtil[aMe]==maxExpUtil){
				optActNum++ ;
				optimalActs[optActNum] = aMe ;
			}
		}
		if (optActNum > -1) {
			int[] optimalActions = new int[optActNum+1];
			for (int i = 0 ; i <= optActNum ; i++ ){
				optimalActions[i] = optimalActs[i] ;
			}
			assert (optimalActs.length <= numActions ) ;
			assert (optimalActs.length == optActNum -1 ) ;
			return optimalActions ;
		}
		else { 
			assert(optimalActs[0] == -1) ;
			return null ;
		}
	}
	
//	/**
//	 * Computes the initial rewards for different actions
//	 * @return i
//	 */
//	public double[][][] getInitialExpRewards() {
//		// type[i][0] = lfType;  // 0 random; 1 follows me; 2 follows other; 3 leader
//		// type[i][1] = gfgType; // 0 generous; 1 fair; 2 greedy
//		// types: 0 = genL, 1 = fairL, 2 = greL, 3 = meF, 4 = otherF, 5 = unknown
//				
//		// Map from types to actions
//		int numTypes = 6 ;
//		int[] typeActs = new int[numTypes] ;  
//		
//		// 0. greedy lead
//		int[] greedyActs = getGreedyActions() ;
//		typeActs[1] = greedyActs[r.nextInt(greedyActs.length)];
//		
//		// 1. fair lead
//		int[] fairActs = getFairActions() ;
//		typeActs[2] = fairActs[r.nextInt(fairActs.length)];
////		int context = fairActs[r.nextInt(fairActs.length)] ; 
//		
//		// 2. generous lead			
//		int[] generousActs = getGenerousActions() ;
//		typeActs[3] = greedyActs[r.nextInt(generousActs.length)];
//		
//		// 3. follow_0			
//		int[] follow_0Acts = optResponseToJ(fairActs[r.nextInt(fairActs.length)], utility) ; 
//		// use a fairAction as an optimistic starting point
//		if (follow_0Acts != null){
//			typeActs[4] = follow_0Acts[r.nextInt(follow_0Acts.length)];
//		} else {
//			int[] newActs = getFairActions() ;
//			typeActs[4] = newActs[r.nextInt(newActs.length)];
//		} 
//		
//		// 4. follow_1
//		int[] follow_1Acts = optResponseToJ(fairActs[r.nextInt(fairActs.length)], utility) ; 
//		// use a fairAction as an optimistic starting point
//		if (follow_1Acts != null){
//			typeActs[5] = follow_1Acts[r.nextInt(follow_1Acts.length)];
//		} else {
//			int[] newActs = getFairActions() ;
//			typeActs[5] = newActs[r.nextInt(newActs.length)];
//		} 
//		
//		// 5. default to safety			
//		typeActs[6] = 0 ; // deal with this in valuation
//		
//		double[][][] initialRewards = new double[numTypes][numTypes][numTypes] ; // type0, type1, myTypeAction
//		
//		for (int t0 = 0 ; t0 < numTypes ; t0++){
//			for (int t1 = 0 ; t1 < numTypes ; t1++){
//				for (int ti = 0 ; ti < numTypes ; ti++){
//		
//					int ai = typeActs[ti];
//					if (ti == 5 ) ai = r.nextInt(numActions) ;
//					int a0 = typeActs[t0];
//					if (t0 == 5 ) a0 = r.nextInt(numActions) ;
//					int a1 = typeActs[t1];
//					if (t1 == 5 ) a1 = r.nextInt(numActions) ;
//					
//					int[] util = utility.getUtility(new int[]{ai,a0,a1}) ;
//					initialRewards[t0][t1][ti] = util[0] ;
//				}
//			}
//		}
//		return initialRewards;
//	}
		
}

/**
 *  A safety strategy to default to if everything goes wrong.
 *  Based on Exp3 under perfect monitoring
 *  @author Archie
 */
class ExpWeighting {
	
	GeneralizedLemonadeUtility utility ;
	Random r = new Random() ;
	
	private int numActions ;
	private double[] weight ;
	private double[] probability ;
	private double gamma ;
	private int gammaT = 1;
	private static final double GAMMA_CONSTANT = 2 ;
	
	protected ExpWeighting(GeneralizedLemonadeUtility utility) {
		numActions = utility.getIslandSize() ;
		weight = initialWeights(utility) ;
		probability = new double[numActions] ;
	}

	private double[] initialWeights(GeneralizedLemonadeUtility utility) {
		double[] actionQuality = new double[numActions] ;
		for(int ai = 0 ; ai < actionQuality.length; ai++ ){
			int oppositeAction = (ai + numActions/2)% numActions ;
			int[] util = utility.getUtility(new int[]{ai,oppositeAction,oppositeAction}) ;
			actionQuality[ai] = util[0];
		}
		return actionQuality ;
	}

	protected void updateExpWeighting(int[] actions, GeneralizedLemonadeUtility utility) {
		gammaT++ ;
		gamma = GAMMA_CONSTANT/gammaT ;
		for (int ai = 0 ; ai < weight.length; ai++ ){
			int util[] = utility.getUtility(new int[]{ai,actions[1],actions[2] }) ;
			weight[ai] *= Math.exp( gamma/numActions * util[0] ) ;
		}
	}

	public int getExpWeightingAction() {
		double sumWeights = 0 ;
		for (int i = 0 ; i < weight.length; i++ ){
			sumWeights += weight[i] ;
		}
		double sumProbs = 0 ;
		double[] cumulativeProbs = new double[numActions] ;
		for (int i = 0 ; i < cumulativeProbs.length; i++ ){
			probability[i] = (1-gamma)* weight[i]/sumWeights + gamma/numActions ;
			sumProbs += probability[i] ;  
			cumulativeProbs[i] = sumProbs ; 
		}		
		double randomLevel = r.nextDouble() * sumProbs;
		int safetyAction = 0;
		for (int i = 0 ; i < cumulativeProbs.length; i++ ){
			if (randomLevel <= cumulativeProbs[i] )
				break ;
			safetyAction++ ;
		}				
		return safetyAction ;
	}
}

///**
// * Counter class for stickCounter, and any other counters
// * @author Archie
// */
//class Counter {
//	public int count;
//
//	public Counter(int initialCount) {count = initialCount;}
//
//	public void setCount(int newCount) {count = newCount;}
//	public void increaseCount() {count++;}
//	public void decreaseCount() {count--;}
//	public int getCount() {return count;}  
//}
