package edu.rutgers.rl3;

import game.Game;
import game.Strategy;

import java.util.Random;
import java.util.Vector;

import util.Array;
import lemonade.GeneralizedLemonadeUtility;

/**
 * Coalition-based agent, but only considering how the outsider 
 * affects us, not our partner. 
 * 
 * The game is divided into two subparts: 
 *  1. The initial step
 *  2. Responding to players given the current actions
 * 
 *  We select our initial action as follows:
 *  Assuming the other players act randomly, what is the best location? 
 *  Given that another player is on the best spot, 
 *  what is the highest valued action? 
 *  
 *  The reasoning is that it is likely that one or both players 
 *  will choose an action close to the best one. 
 *  It is also likely that given two players are on the "best" 
 *  and "next-best" spots, the third player would prefer picking a 
 *  place near the best, leaving the next-best player with 
 *  a net higher score as the two others fight over the best place. 
 *  
 *  The decision of when to move from this initial spot 
 *  (or any current location) is calculated by keeping a tally of 
 *  how many points the number one player is receiving, 
 *  and moving once these points reaches a threshold. 
 *  This tally is also adjusted for situations where moving would 
 *  be a worse decision than letting another player move. 
 *  
 *  The decision of where to move is narrowed down by looking for 
 *  current coalitions, as defined by both players in a coalition 
 *  performing better than the third player.  However, 
 *  this iteration only ensures that our agent does better than the 
 *  third (keeping our partner fixed), leaving open the possibility 
 *  that our partner does worse.  We reason that the risk of 
 *  moving should be worth this extra bonus.  
 *  
 *  
 * @author wunder
 */
public class JerseyBeachcomber implements Strategy{
  /**
   * The action the player plays
   */
  int constantAction;
  int turn = 0;

  int[] bestActions;
  int[] nextBestActions;
  
  int[] prevActions = new int[3];
  int[][] initialCoalitions;
  int[] availableCoalitionActions;
  int[] densities;
  int totalUtility = 0;
  int numActions;
  int[] scores = new int[3];
  
  boolean unchanged = false; 
  public final static int DEBUG = 0;

  public final static int LOSS_THRESHOLD = 200;
  public final static int SINGLE_TURN_THRESHOLD = 200;
  
  int thresholdShift = 0;
  
  GeneralizedLemonadeUtility utility;
  //HighDensityConstantNB hdcNB;
  //HighDensityConstant hdc;

  /**
   * Construct a coalition strategy. 
   * The density vector is used to project coalitions 
   * onto the space, which can then be used to 
   * narrow down actions and find acceptable new spots. 
   * 
   * @param g
   * @param seed
   */
  public JerseyBeachcomber(Game g, long seed, String options) {
	  
    this.utility = (GeneralizedLemonadeUtility)g.getUtility();
    this.densities = utility.getIslandDensity();
    numActions = g.getNumActions();
    availableCoalitionActions = new int[numActions];
    Random r = new Random(seed);

    initBestActions(seed);
    constantAction = nextBestActions[r.nextInt(nextBestActions.length)];

    for (int i = 0; i<densities.length; i++) 
	    totalUtility += densities[i]; 
  }
  
  protected int getQuality(int action){
	    int totalValue = 0;
	    int oppositeAction = (action+utility.getIslandSize()/2)% utility.getIslandSize();
	    int[] util = utility.getUtility(new int[]{action,oppositeAction,oppositeAction});
	    for (int i = 0; i<utility.getIslandSize(); i++) {
		for (int j = 0; j<utility.getIslandSize(); j++) {
			util = utility.getUtility(new int[]{action,i,j});
			totalValue += util[0];
		}
	    }
	    return totalValue;
	  }

	  protected int getNextBestQuality(int action, int otherAction) {
		int totalValue = 0;
	    int[] util;
	    for (int i = 0; i<utility.getIslandSize(); i++) {
			util = utility.getUtility(new int[]{action,i,otherAction});
			totalValue += util[0];
	    }
	    return totalValue;
	  }
	  
	  private int[] initBestActions(long seed) {
	    Random r = new Random(seed);
	    int[] quality = new int[utility.getIslandSize()];
	    int[] qualityNB = new int[utility.getIslandSize()];
	    for(int i=0;i<quality.length;i++){
	      quality[i]=getQuality(i);
	    }
	    int maxDensity = quality[0];
	    int maxDensityCount = 1;
	    for(int i=0;i<quality.length;i++){
	      if (maxDensity<quality[i]){
	        maxDensity=quality[i];
	        maxDensityCount=1;
	      } else if (maxDensity==quality[i]){
	        maxDensityCount++;
	      }
	    }
	    bestActions = new int[maxDensityCount];
	    int bestActionIndex = 0;
	    for(int i=0;i<quality.length;i++){
	      if (quality[i]==maxDensity){
	        bestActions[bestActionIndex]=i;
	        bestActionIndex++;
	      }
	    }

	    //int bestAct = bestActions[r.nextInt(bestActions.length)];
	    for (int b = 0; b<bestActions.length; b++) {
	    	for(int i=0;i<qualityNB.length;i++){
	    		qualityNB[i]=getNextBestQuality(i,bestActions[b]);
	    	}
	    }
	    maxDensity = qualityNB[0];
	    maxDensityCount = 1;
	    for(int i=0;i<qualityNB.length;i++){
	      if (maxDensity<qualityNB[i]){
	        maxDensity=qualityNB[i];
	        maxDensityCount=1;
	      } else if (maxDensity==qualityNB[i]){
	        maxDensityCount++;
	      }
	    }
	    nextBestActions = new int[maxDensityCount];
	    int nextBestActionIndex = 0;
	    for(int i=0;i<quality.length;i++){
	      if (qualityNB[i]==maxDensity){
	        nextBestActions[nextBestActionIndex]=i;
	        nextBestActionIndex++;
	      }
	    }
	    return quality; 
	  }
	  
	  protected int qualityRange(int[] qual) {
		  int total = 0;
		  int max = 0;
		  for (int i=0; i<qual.length; i++) {
			  total+=qual[i];
			  if (max>qual[i]) 
				  max = qual[i];
		  }
		  return max - total/numActions;		  
	  }
	  
  // Are X and Y in a "coalition" against Z? 
  // Do they get higher scores than Z and 
  // relatively equal to each other? 
  protected boolean coalitionTest(int[] util) {
	return (!(util[2] > util[0]));
	//return (!((util[2] > util[0]) || (util[2] > util[1])));
  }
  
  //return an array of the possible coaltion-building spots. 
  //0 = none, 1 = coalition with player 1, 2 = coalition w/P2, 3 = BOTH
  public int[] currentCoalitions(int[] actions, int[] coalitions) {
	//int[] coalitions = new int[numActions];
	boolean coalitionI = true; 
	boolean coalitionJ = true;
	int[] origUtil = utility.getUtility(new int[]{actions[0],actions[1],actions[2]});
	for (int i = 0; i< numActions; i++) {
	   coalitionI = true;
  	   coalitionJ = true;
	   
 	   int[] util = utility.getUtility(new int[]{i,actions[1],actions[2]});
	   int[] util1;
	   int[] util2;
		int maxScoreI = util[2];
		int maxScoreJ = util[1];

		for (int k = 0; k<numActions; k++) {
			util1 = utility.getUtility(new int[]{i,actions[1],k});
			util2 = utility.getUtility(new int[]{i,actions[2],k});
			if (util1[2] > maxScoreI) 
				maxScoreI = util1[2];
			if (util2[1] > maxScoreJ) 
				maxScoreJ = util2[1]; 
		}
		for (int k = 0; k<numActions; k++) {
			util1 = utility.getUtility(new int[]{i,actions[1],k});
			util2 = utility.getUtility(new int[]{i,actions[2],k});
			coalitionI = (coalitionI&&coalitionTest(util1));//new int[]{i,actions[1],k}));
			coalitionJ = (coalitionJ&&coalitionTest(util2));//new int[]{i,actions[2],k}));
		}
	   
	   if (coalitionI) 
		   coalitions[i]++;
	   if (coalitionJ)
		   coalitions[i]+=2;
	   }
	return coalitions;
  }


  public int getAction() {
    int[] util = utility.getUtility(prevActions);
    int bestValue = util[0];
    int bestOther = 0;
    int bestAction = constantAction; 
    int[] utilC = new int[3];
    int[] utilAlt = new int[3];
    
    for (int i = 0; i< 3; i++) 
		scores[i]+=util[i];
    
    if (scores[0]>scores[1] && scores[0]>scores[2]) {
    	if (unchanged)
    		return constantAction;
    	else {
    		for (int i = 0; i< 3; i++) 
    			scores[i] = 0;
    		thresholdShift = 0;
    	}
    }
    
    if (!unchanged) {
    	for (int i = 0; i<availableCoalitionActions.length; i++) {
    		availableCoalitionActions[i] = 0;
    	}
            currentCoalitions(prevActions,availableCoalitionActions);
    }
    for (int i = 0; i<availableCoalitionActions.length; i++) {
    	if (availableCoalitionActions[i]>0) {
    		utilC = utility.getUtility(new int[]{i,prevActions[1],prevActions[2]});
    		if (utilC[0]>bestValue) {
    			bestValue = utilC[0];
    			bestAction = i;
    		}
    	}
    }
    
    int sucker = 0;
    if (util[1]<=util[2])
    	sucker = 1;
    else 
    	sucker = 2;
    int otherBest = 0;
    int otherValue = 0;
    for (int i = 0; i<numActions; i++) {
    		utilAlt = utility.getUtility(new int[]{i,prevActions[3-sucker],prevActions[0]});
    		if (utilAlt[0]>otherValue) {
    			otherValue = utilAlt[0];
    			otherBest = i;
    		}
    }
    utilAlt = utility.getUtility(new int[]{prevActions[0],prevActions[3-sucker],otherBest});
    thresholdShift += bestValue - utilAlt[0]; 
    
    int max = scores[0];
    if (scores[1]>max) {
    	if (scores[2]>scores[1])
    		max = scores[2];
    	else
    		max = scores[1];
    }
    else {
    	if (scores[2]>max)
    		max = scores[2];
    }
    
//  if ((max - LOSS_THRESHOLD > util[0] ||util[2] - LOSS_THRESHOLD > util[0])&& bestValue > util[0] && turn > 2) {
//    if (totalUtility/3 > util[0] && bestValue > util[0] && turn > 1) {
    if ((max - scores[0] + thresholdShift > LOSS_THRESHOLD)&& bestValue > util[0] && turn > 1) {    	
	    constantAction = bestAction; 
	    for (int i = 0; i< 3; i++) 
			scores[i]=0;
		thresholdShift = 0;
    }
    // We haven't found a good coalition, but let's move anyway.
    // we are on a three-way tie so might as well look for a better spot.
    else if ((max - totalUtility/3 <= 1 || max + thresholdShift - scores[0] > SINGLE_TURN_THRESHOLD*3) && turn > 3) {
    	int bestA = bestPayoffs(currentPayoffs(prevActions));
    	int[] bestU = utility.getUtility(new int[]{bestA,prevActions[1],prevActions[2]});
    	if (bestU[0] - SINGLE_TURN_THRESHOLD >= util[0] || (max + thresholdShift - SINGLE_TURN_THRESHOLD*3 > scores[0] && bestU[0] > util[0]))
    		constantAction = bestA; 
    	for (int i = 0; i < 3; i++) 
    		scores[i]=0;
		thresholdShift = 0;
	}
	return constantAction;
  }

  public void observeOutcome(int[] actions) {
	  unchanged = (actions[0]==prevActions[0]&&actions[1]==prevActions[1]&&actions[2]==prevActions[2]);
	  turn++;
	  for (int i=0; i<actions.length; i++) 
		  prevActions[i] = actions[i];
	  if (DEBUG==-1)
		  printGame(actions);
  }
  
  // find the payoffs of a given P0/P1 arrangement, with P2 playing randomly
  protected int[] uniformPayoffs(int[] actions) {
	  int[] util;
	  int[] totalUtil = new int[numActions];
	  for (int i = 0; i<numActions; i++) {
		  for (int k = 0; k<numActions; k++) {
	  		  util= utility.getUtility(new int[]{i,actions[1],k});
  		  	  totalUtil[i] += util[0]; 
	      }
		  totalUtil[i] = totalUtil[i]/12;
	  }
	  return totalUtil; 
  }
  
  protected void printGame(int[] actions) {
	  String beach = "";
	  for (int i = 0; i<numActions; i++) {
		  if (actions[0]==i)
			  beach = beach + "A";
		  else beach = beach + " ";
	  	  if (actions[1]==i)
			  beach = beach + "B";
	  	  else beach = beach + " ";
	  	  if (actions[2]==i)
			  beach = beach + "C";
	  	  else beach = beach + " ";
	  }
	  int[] util = utility.getUtility(actions);
	  System.out.println(beach + Array.arrayToNakedString(util));
  }
  
  protected int bestPayoffs(int[] util) {
	  int maxV = 0;
	  int maxA = 0;
	  for (int i = 0; i<util.length; i++) {
		  if (maxV > util[i]) {
			  maxV = util[i];
			  maxA = i;
		  }
		  if (maxV == util[i] && (maxV/2)%2==1) {
			  maxV = util[i];
			  maxA = i;
		  }
	  }
	  return maxA;
  }
  
  protected int[] currentPayoffs(int[] actions) {
	  int[] util;
	  int[] totalUtil = new int[numActions];
	  for (int i = 0; i<numActions; i++) {
	  		  util = utility.getUtility(new int[]{i,actions[1],actions[2]});
  		  	  totalUtil[i] = util[0]; 
	  }
	  return totalUtil; 
  }
  
}
